/*
m_3d_obstacle - emulate obstacles between soource and a listener in a room

DESCRIPTION
m_3d_obstacle emulates the effect of obstacles in a room. When an obstacle is
between the listener and the audio source, an effect can be heard.
NOTE: This is NOT a true model. A true model would need to calculate early
reflections and wave propagation.

SYNTAX
aleft, aright m_3d_obstacle arc, ksrcx, ksrcy, ksrcz, klistenerx, klistenery, \
	klistenerz, klistener_headrot, iobstaclefn, iroomfn

INITIALIZATION
iobstaclefn - ftable describing all obstacles, the format is:
	x = x coordinate of the obstacle in metres
	y = y coordinate of the obstacle in metres
	z = height of the obstacle in metres
	width = width and breadth of the obstacle in metres
	proximity = distance in which the obstacle starts to influence the sound
iroomfn - ftable describing a room (see m_3d_atmo for the format)

PERFORMANNCE
arc - audio source signal
ksrcx, ksrcy, ksrcz - x, y and z coordinate of the audio source in metres
klistenerx, klistenery, klistenerz - x, y and z coordinate of the listener
	in metres
klistener_headrot - Head rotation of the listener in degrees (0-360)

CREDITS
Author: Jeanette C., many thanks to the Csound mailinglist and Brian Carty,
author of the hrtf opcodes.
*/

opcode m_3d_obstacle, aa, akkkkkkkii
	asrc, ksrcx, ksrcy, ksrcz, klx, kly, klz, kheadrot, iobstaclefn, iroomfn xin

	; Read the room parameters
	irfade table 0, iroomfn, 0
	irorder table 1, iroomfn, 0
	irx table 2, iroomfn, 0
	iry table 3, iroomfn, 0
	irz table 4, iroomfn, 0
	irwallhi table 5, iroomfn, 0
	irwalllo table 6, iroomfn, 0
	irwallg1 table 7, iroomfn, 0
	irwallg2 table 8, iroomfn, 0
	irwallg3 table 9, iroomfn, 0
	irfloorhi table 10, iroomfn, 0
	irfloorlo table 11, iroomfn, 0
	irfloorg1 table 12, iroomfn, 0
	irfloorg2 table 13, iroomfn, 0
	irfloorg3 table 14, iroomfn, 0
	irceilhi table 15, iroomfn, 0
	irceillo table 16, iroomfn, 0
	irceilg1 table 17, iroomfn, 0
	irceilg2 table 18, iroomfn, 0
	irceilg3 table 19, iroomfn, 0
	Sleft sprintf "hrtf-%d-left.dat", sr
	Sright sprintf "hrtf-%d-right.dat", sr

	; Initialise variables for the comparison
	koldx init -1 ; lst position of the source x
	koldy init -1 ; last y
	koldz init -1 ; last z
	kndx init 0 ; Index of the obstacle to compare to
	amodsrc init 0
	kxc init 0
	kyc init 0
	inum_obstacles = ftlen(iobstaclefn) / 5
	kox init 0 ; obstacle x coordinate
	koy init 0 ; obstacle y coordinate
	koz init 0 ; obstacle z coordinate
	kosize init 0 ; obstacle size widt/breadth
	koprox init 0 ; proximity of effect
	kothresh init 0 ; overall thresshold when effect starts
	kcross init 0
	kxl init 0
	kyl init 0
	koldist init 0 ; distance current obstacle to listener
	kfactor init 1 ; Factor to affect the signal, 1 nothing, 0 full effect
	kmodf init 1

	; Determine whether an obstacle is between source and listener.
	klsdist = sqrt((klx - ksrcx)^2 + (kly - ksrcy)^2) ; distance listener to
		; source on the x-y plane
	if (ksrcx != koldx) || (ksrcy != koldy) || (ksrcz != koldz) then ; start new comparison
		kndx = 0
		while (kndx < inum_obstacles) do
			kox table (kndx * 5), iobstaclefn, 0
			koy table (kndx * 5) +1, iobstaclefn, 0
			koz table (kndx * 5) +2, iobstaclefn, 0
			koldist = sqrt((kox - klx)^2 + (koy - kly)^2)
			if (koldist < klsdist) && (koz >= ksrcz) then ; Distance of ostacle to
				; listener is shorter than src listener and obstacle is higher
				; than the source
				kosize table (kndx * 5) +3, iobstaclefn, 0
				koprox table (kndx * 5) +4, iobstaclefn, 0
				kothresh = (kosize / 2) + koprox
				kxc = kox - ksrcx
				kyc = koy - ksrcy
				kxl = klx - ksrcx
				kyl = kly - ksrcy
				kcross = (kxc * kyl) - (kyc * kxl)
				if abs(kcross) < kothresh then ; We're in the proximity area?
					if abs(kxl) >= abs(kyl) then
						if kxl > 0 then
							if (ksrcx <= kox) && (kox <= klx) then
								kfactor limit ((abs(kcross) - kosize /2) / kothresh), 0, 1
								kndx = inum_obstacles
							else
								kfactor = 1
							endif
						else
							if (klx <= kox) && (kox <= ksrcx) then
								kfactor limit ((abs(kcross) - kosize /2) / kothresh), 0, 1
								kndx = inum_obstacles
							else
								kfactor = 1
							endif
						endif
					else
						if kyl >0 then
							if (ksrcy <= koy) && (koy <= kly) then
								kfactor limit ((abs(kcross) - kosize /2) / kothresh), 0, 1
								kndx = inum_obstacles
							else
								kfactor = 1
							endif
						else
							if (kly <= koy) && (koy <= ksrcy) then
								kfactor limit ((abs(kcross) - kosize /2) / kothresh), 0, 1
								kndx = inum_obstacles
							else
								kfactor = 1
							endif
						endif
					endif
				else
					kfactor = 1
				endif
			endif
			kndx += 1
		od
	endif
	koldx = ksrcx
	koldy = ksrcy
	koldz = ksrcz

	; modify the source signal
	if kfactor != 1 then ; source must be actively modified
		kmodf logcurve kfactor, 3
		kmodf = (kmodf * 2 + 1) / 3
		amodsrc butterlp asrc*kmodf, (sr / 2) * kmodf
	else ; modified source is identical to original source
		amodsrc = asrc
	endif

	; Process the modified audio source and output
	aleft, aright, irtl1, irth1, imfp1 hrtfearly amodsrc, ksrcx, ksrcy, ksrcz, klx, kly, klz, Sleft, Sright, 0, irfade, sr, irorder, 1, kheadrot, irx, iry, irz, irwallhi, irwalllo, irwallg1, irwallg2, irwallg3, irfloorhi, irfloorlo, irfloorg1, irfloorg2, irfloorg3, irceilhi, irceillo, irceilg1, irceilg2, irceilg3
	xout aleft, aright
endop
